#ifndef GENERATOR_FRAMEDATA_HH
#define GENERATOR_FRAMEDATA_HH

#include "DataSource.hh"
#include "Dacc.hh"
#include <list>

namespace generator {

  /**  %FrameData allows the user to inject pre-recorded data into the 
    *  generation streams. Each %FrameData source has uses the following 
    *  parameters:
    *  <table>
    *  <tr><td>%Parameter</td>
    *      <td>Contents</td>
    *  </tr>
    *  <tr><td>%Channel</td>
    *      <td>Name of channel providing the data.</td>
    *  </tr>
    *  <tr><td>Files</td>
    *      <td>Path(s) to one or more (globbed) frame files.</td>
    *  </tr>
    *  </table>
    *
    *  \note The data source sample rate must be the same as 
    *  the sample rate defined in the frame data.
    *
    *  @memo Data source reading channels from frame files.
    *  @author John Zweizig
    *  @version 1.1; last modified April 13, 2004
    */
  class FrameData : public DataSource {
  public:
      using DataSource::setParameter;

  public:
      /** Frame data constructor.
        */
      FrameData(void);

      /**  Construct a frame data source and initialize the frame file name
        *  and channel name parameters.
        *  @memo Parameter constructor.
        *  @param files   A list of files to be searched for data.
	*  @param channel name of channel to be fetched.
        */
      FrameData(const char* files, const char* channel);

      /**  Pre-recorded data source destructor.
        *  @memo Destructor.
	*/
      ~FrameData(void);

      /**  Make an identical copy of a frame data source.
        *  @memo Clone a frame data source.
	*  @return Pointer to the newly created object.
        */
      FrameData* clone(void) const;

      /**  Return the data source type name ("FrameData").
        *  @memo Source data type name string.
	*  @return Constant pointer to source type name string.
        */
      const char* getDataType(void) const;

      /**  Specify the name of the channel to be read in.
        *  @memo Set channel name
	*  @param name Channel name string.
        */
      void setChannel(const std::string& name);

      /**  Specify the frame file(s) to be searched for appropriate data. The
        *  frame file(s) may be specified as a single frame path or as a 
	*  globbed path specifier.
        *  @memo Set the file list.
	*  @param frames Frame name list.
        */
      void setFileList(const std::string& frames);

      /**  Specify the value of a parameter
        *  @memo Set a parameter 
	*  @param name Parameter name
	*  @param val  Parameter value.
	*/
      virtual void setParameter(const std::string& name, 
				const std::string& val);

  private:  // private data types
      /**  The data input class maintains a list of files to be searched by 
        *  a frame data source. The input data structures may be shared by 
	*  several %FrameData data sources for efficient access of multiple 
	*  data channels.
        *  @memo Data input class.
	*/
      class input {
      public:
	/**  Construct a data input object and optionally set the file list 
	  *  parameter.
	  */
	input(const char* fl=0);

	/**  Destroy the data input object.
	 */
	~input(void);

	/**  Get the File list specifier.
	  */
	const char* getFileList(void) const;

	/**  Set the file list.
	  */
	void setFileList(const std::string& fList);

	/**  Get the specified data string.
	  */
	void getData(const char* ch, const Time& t0, Interval Dt, 
		     int N, float* data);

	/**  Return a referece to the internal data accessor.
	  */
	Dacc& refDacc(void);
	const Dacc& refDacc(void) const;

      private:
	std::string mFileList;
	Dacc        mIn;
      };
      typedef std::list<input> InpList;
      typedef InpList::iterator  inp_iter;

  private:
      /**  Get a data series for the specified Time interval.
        */
      void getSeries(const Time& t0, Interval dT, int N, float* data);

      /**  Search the input accessor list for one covering the correct data 
        *  files.
	*/
      inp_iter findInput(const std::string& filelist);

      /**  Add a new input accessor to the list.
	*/
      inp_iter addInput(const std::string& filelist);

  private:  // private data
      static InpList mInputList;
      std::string  mChannel;
      inp_iter mInput;
  };


  //====================================  In-line FrameData methods.
  inline const char*
  FrameData::getDataType(void) const {
      return "FrameData";
  }

  //====================================  In-line Framedata::input methods.
  inline Dacc& 
  FrameData::input::refDacc(void) {
      return mIn;
  }

  inline const Dacc& 
  FrameData::input::refDacc(void) const {
      return mIn;
  }

  inline const char* 
  FrameData::input::getFileList(void) const {
      return mFileList.c_str();
  }

} // namespace generator

#endif  // GENERATOR_FRAMEDATA_HH
