static const char *versionId = "Version $Id: timeseries.cc 6416 2011-04-01 19:09:15Z james.batch@LIGO.ORG $" ; 
/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: timeseries						*/
/*                                                         		*/
/* Module Description: triggered time series measurement		*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

/* Header File List: */

#include <time.h>
#include <math.h>
#include  <iostream>
#include "dtt/diagnames.h"
#include "dtt/timeseries.hh"
#include "dtt/awgfunc.h"
#include "gdsalgorithm.h"
#include "dtt/diagdatum.hh"
#include "IIRFilter.hh"
#include "FilterDesign.hh"

namespace diag {
   using namespace std;
   using namespace thread;


   // maximum number of sampling points per (excitation) period
   const double _MAX_OVERSAMPLING = 4.0;
   // Time ahead buffers are allocated
   const double _TIME_AHEAD = 3.0; 

   timeseries::tmpresult::tmpresult (string Name, int Size, 
                     double Dx, bool Cmplx, string Filter) 
   : name (Name), cmplx (Cmplx), size (Size), x (0), xsqr (0), dx (Dx),
   filterspec (Filter), filter (0)
   {
      allocate (Size);
   }


   timeseries::tmpresult::~tmpresult()
   {
      allocate ();
      if (filter) delete filter;
   }


   bool timeseries::tmpresult::allocate (int Size)
   {
      // delete old arrays
      if (xsqr != 0) {
         delete [] xsqr;
         xsqr = 0;
      }
      if (x != 0) {
         delete [] x;
         x = 0;
      }
      // alocate new ones
      size = Size;
      if (size == 0) {
         return true;
      }
      else {
         x = new (nothrow) float[cmplx ? 2 * size : size];
         xsqr = new (nothrow) float[cmplx ? 2 * size : size];
         return valid();
      }
   }


   bool timeseries::tmpresult::valid () const
   {
      return ((x != 0) && (xsqr != 0) && (dx > 0));
   }


   timeseries::tmpresult::tmpresult (const tmpresult& tmp) 
   : name (""), cmplx (false), size (0), x (0), xsqr (0), dx (1),
   filter (0) {
      *this = tmp;
   }


   timeseries::tmpresult& timeseries::tmpresult::operator= (
                     const tmpresult& tmp)
   {
      if (this != &tmp) {
         name = tmp.name;
         size = tmp.size;
         cmplx = tmp.cmplx;
         x = tmp.x;
         xsqr = tmp.xsqr;
         dx = tmp.dx;
         const_cast <tmpresult&>(tmp).x = 0;
         const_cast <tmpresult&>(tmp).xsqr = 0;
         filterspec = tmp.filterspec;
         if (filter) delete filter;
         filter = 0;
      }
      return *this;
   }


   timeseries::timeseries () 
   : stdtest (timeseriesname) , measTime (0), preTriggerTime (0),
   settlingTime (0), sigBW (0)
   {
   }


   diagtest* timeseries::self () const 
   {
      return new (nothrow) timeseries ();
   }


   bool timeseries::end (ostringstream& errmsg)
   
   {
      semlock		lockit (mux);
      // delete temporary storage
      tmps.clear();
   
      return stdtest::end (errmsg);
   }


   bool timeseries::readParam (ostringstream& errmsg)
   {
      // call parent method
      if (!stdtest::readParam (errmsg)) {
         return false;
      }
   
      semlock		lockit (mux);
      bool		err = false;
   
      // read measurement time
      if (!test->getParam (*storage->Test,
                          tsMeasurementTime, measTime)) {
         errmsg << "Unable to load values from Test." <<
            tsMeasurementTime << endl;
         err = true;
      }
      // read pre trigger time
      if (!test->getParam (*storage->Test,
                          tsPreTriggerTime, preTriggerTime)) {
         errmsg << "Unable to load value from Test." <<
            tsPreTriggerTime << endl;
         err = true;
      }
      // read settling time
      if (!test->getParam (*storage->Test,
                          tsSettlingTime, settlingTime)) {
         errmsg << "Unable to load values from Test." <<
            tsSettlingTime << endl;
         err = true;
      }
      // read settling time
      if (!test->getParam (*storage->Test, tsDeadTime, deadTime)) {
         errmsg << "Unable to load values from Test." <<
            tsDeadTime << endl;
         err = true;
      }
      // read signal bandwidth
      if (!test->getParam (*storage->Test, tsBW, sigBW)) {
         errmsg << "Unable to load value from Test." << tsBW << endl;
         err = true;
      }
      // read whether to include statistics
      if (!test->getParam (*storage->Test, tsIncludeStatistics, 
                          includeStatistics)) {
         errmsg << "Unable to load value from Test." << 
            tsIncludeStatistics << endl;
         err = true;
      }
      // read filter design string
      if (!test->getParam (*storage->Test, tsFilter, filterSpec)) {
         filterSpec = "";
      }
   
      // read stimuli channel
      if (!readStimuliParam (errmsg, false, allWaveforms)) {
         err = true;
      }
   
      // read measurement channels
      if (!readMeasParam (errmsg)) {
         err = true;
      }
   
      // check heterodyne frequency
      fZoom = 0.0;
      if (!heterodyneFrequency (fZoom)) {
         errmsg << "Channels have inconsistent heterodyne frequency." << endl;
         err = true;
      }
      //cout << "ZOOM FREQUENCY IN TIME SERIES MEASUREMENT " << fZoom << endl;
   
      return !err;
   }


   bool timeseries::calcTimes (std::ostringstream& errmsg,
                     tainsec_t& t0)
   {
      semlock		lockit (mux);
      bool		err = false;
   
      // check measurement time
      if (measTime <= 0) {
         errmsg << "Measurement time must be positive" << endl;
         return false;
      }
      // check settling time
      if (settlingTime < 0) {
         errmsg << "Settling time must be positive or zero" << endl;
         return false;
      }
      // check dead time
      if (deadTime >= 1) {
         errmsg << "Dead time must be smaller than one" << endl;
         return false;
      }
      // check pre-trigger time
      if (preTriggerTime >= 1) {
         errmsg << "Pre-trigger time must be smaller than one" << endl;
         return false;
      }
      // check averages
      if (averages < 1) {
         errmsg << "Number of averages must be at least one" << endl;
         err = true;
      }
      // check averages
      if (averages > 100000) {
         errmsg << "Number of averages must be smaller than 100000" << 
            endl;
         err = true;
      }
      // check average type
      if ((averageType < 0) || (averageType > 2)) {
         errmsg << "Illegal average type" << endl;
         err = true;
      }
      // check bandwidth
      if (sigBW <= 0) {
         errmsg << "Bandwidth must be positive" << endl;
         err = true;
      }
      // check filter
      FilterDesign ds (16384.);
      if (!ds.filter (filterSpec.c_str())) {
         errmsg << "Invalid filter" << endl;
         err = true;
      }
      if (err) {
         return false;
      }
      rnumber = 0;
      for (stimuluslist::iterator iter = stimuli.begin();
          iter != stimuli.end(); iter++) {
         if (iter->isReadback) {
            rnumber++;
         }
      }
      rnumber += meas.size();
   
      // highest frequency of interest
      fMaxMeas = sigBW;
      // determine sampling rates
      samplingFrequencies (fMaxMeas, fMinSample, fMaxSample);
   
      // calculate time grid
      timeGrid = calcTimeGrid (fMaxSample / 2.0, &t0);
   
      // calculate measurement start time
      if (preTriggerTime >= 0) {
         // triggere is applied after measurement starts
         mStart = adjustForSampling (settlingTime * measTime, timeGrid);
      }
      else {
         // trigger is applied before measurement starts
         mStart = adjustForSampling (settlingTime * measTime - 
                              preTriggerTime * measTime, timeGrid);
      }
      exct0 = mStart + preTriggerTime * measTime;
      mTimeAdd = adjustForSampling (measTime, timeGrid) - measTime;
      if (dTime >= 0) {
         dTime = adjustForSampling (deadTime, timeGrid);
      }
      else {
         dTime = - adjustForSampling (-deadTime, timeGrid);
      }
   
      // handle averaging
      avrgsize = averages;
      if (avrgsize * (measTime + mTimeAdd) < _TIME_AHEAD) {
         // need more buffers
         avrgsize = (int)(_TIME_AHEAD / (measTime + mTimeAdd));
      }
      // make sure we account for filter delay
      int onum = (int)(ceil ((measTime + mTimeAdd + 21./fMinSample) / 
                            (measTime + mTimeAdd + 1E-12)) + 0.1);
      if (onum > 1) avrgsize += onum;
      avrgsize += 1;
      // if (avrgsize < onum) {
         // avrgsize = onum;
      // }
      cout << "AVRGSIZE IS _________________" << avrgsize << endl;
   
      return true;
   }


   bool timeseries::newMeasPoint (int i, int measPoint)
   {
      semlock		lockit (mux);
   
      // calulate start time
      tainsec_t	start = T0 + (tainsec_t)
         ((mStart + (double) (i + skipMeas) * 
          (measTime + mTimeAdd)) * 1E9 + 0.5);
   
      // check if too far behind
      if (RTmode) {
         tainsec_t now = TAInow();
         if (start < now + _EPOCH) {
            skipMeas = 
               (int) (((double) (now + _EPOCH - T0) / 1E9 - mStart) /
                     (measTime + mTimeAdd) + 0.99) - i;
            if (skipMeas < 0) {
               skipMeas = 0;
            }
            start = T0 + (tainsec_t)
               ((mStart + (double) (i + skipMeas) * 
                (measTime + mTimeAdd)) * 1E9 + 0.5);
         }
      }
      cerr << "start " << start << " " << mStart << " " <<
         (measTime + mTimeAdd) << endl;
   
      // fine adjust to time grid & calc. duration
      start = fineAdjustForSampling (start, timeGrid);
      tainsec_t	duration = 
         (tainsec_t) ((measTime + mTimeAdd - dTime) * 1E9 + 0.5);
   
      // add interval
      intervals.push_back (interval (start, duration));
   
      // add new partitions
      if (!addMeasPartitions (intervals.back(), measPoint * averages + i,
                           fMaxSample, 0, fZoom)) {
         return false;
      }
   
      // add synchronization point
      if (!addSyncPoint (intervals.back(), i, measPoint)) {
         return false;
      }
   
      return true;
   }


   bool timeseries::calcMeasurements (std::ostringstream& errmsg,
                     tainsec_t t0, int measPoint)
   {
      semlock		lockit (mux);
      tainsec_t		eStart;		// excitation start	
      tainsec_t		dur;		// excitation duration	
   
      // determine excitation signals
      eStart = t0 + (tainsec_t) (exct0 * 1E9);
      for (stimuluslist::iterator iter = stimuli.begin();
          iter != stimuli.end(); iter++) {
         if ((iter->waveform == (AWG_WaveType) 10) ||
            (iter->waveform == (AWG_WaveType) 11)) {
            dur = (tainsec_t) ((measTime + mTimeAdd) * 1E9 + 0.5);
         }
         else {
            dur = -1;
         }
         double fold = iter->freq;
         if (iter->freq <= 0) {
            iter->freq = 1.0 / (measTime + mTimeAdd);
         }
         if (!iter->calcSignal (eStart, dur)) {
            errmsg << "Unable to calculate excitation signal" << endl;
            iter->freq = fold;
            return false;
         }
         iter->freq = fold;
      }
   
      // create initial measurement points
      skipMeas = 0;
      for (int i = 0; i < avrgsize; i++) {
         if (!newMeasPoint (i)) {
            errmsg << "Unable to create measuremennt points" << endl;
            return false;
         }
      }
      return true;
   }


   bool timeseries::stopMeasurements (int firstIndex)
   {
      semlock		lockit (mux);
      // delete temporary storage
      tmps.clear();
   
      return stdtest::stopMeasurements (firstIndex);
   }


   bool timeseries::analyze (const callbackarg& id, int measnum, 
                     bool& note)
   {
      semlock		lockit (mux);
   
      /////////////////////////////////////////////////////////
      // Init analysis first time around		     //
      /////////////////////////////////////////////////////////
   
      // init analysis first time around
      if (measnum == 0) {
      
      /////////////////////////// Allocate temporary storage
         tmps.clear();
         const diagResult& aChn = diagChn::self();
         int num = id.measPoint * averages + id.measPeriod;
         for (stimuluslist::iterator iter = stimuli.begin();
             iter != stimuli.end(); iter++) {
            if (iter->duplicate) {
               continue;
            }
            if (iter->isReadback) {
               // get time series
               string pname = diagObjectName::makeName (
                                    iter->readbackInfo.chName, step, num);
               gdsDataObject* chndat = storage->findData (pname);
               if (chndat == 0) {
                  return false;
               }
               int N;
               if (!aChn.getParam (*chndat, stTimeSeriesN, N)) {
                  return false;
               }
               double dt;
               if (!aChn.getParam (*chndat, stTimeSeriesdt, dt)) {
                  return false;
               }
               double f0 = 0.0;
               aChn.getParam (*chndat, stTimeSeriesf0, f0);
               // create temp object
               tmps.push_back (tmpresult (iter->readback, N, dt, f0 != 0,
                                         filterSpec));
               if (!tmps.back().valid()) {
                  return false;
               }
            }
         }
         for (measlist::iterator iter = meas.begin();
             iter != meas.end(); iter++) {
            if (iter->duplicate) {
               continue;
            }
            // get time series
            string pname = diagObjectName::makeName (
                                 iter->info.chName, step, num);
            gdsDataObject* chndat = storage->findData (pname);
            if (chndat == 0) {
               return false;
            }
            int N;
            if (!aChn.getParam (*chndat, stTimeSeriesN, N)) {
               return false;
            }
            double dt;
            if (!aChn.getParam (*chndat, stTimeSeriesdt, dt)) {
               return false;
            }
            double f0 = 0.0;
            aChn.getParam (*chndat, stTimeSeriesf0, f0);
            // create temp object
            tmps.push_back (tmpresult (iter->name, N, dt, f0 != 0,
                                      filterSpec));
            if (!tmps.back().valid()) {
               return false;
            }
         }
      
      /////////////////////////// Create index
         // get index access class
         const diagIndex& 	indx = diagIndex::self();
         // find index / create new index if necessary
         gdsDataObject* 	iobj = storage->findData (stIndex);
         if (iobj == 0) {
            iobj = indx.newObject (0);
            if (iobj == 0) {
               return false;
            }
            storage->addData (*iobj, false);
         }
      
         // write index 
         ostringstream		entry1;
         cerr << "ANALYZE TS 1" << endl;
         // time series entry
         int b = 0;
         for (tmpresults::iterator iter = tmps.begin();
             iter != tmps.end(); iter++, b++) {
               // set index entry
            diagIndex::channelEntry (entry1, b, iter->name);
            diagIndex::resultEntry (entry1, b, 0, iter->size, b);
         }
         indx.setEntry (*iobj, icTimeseries, step, entry1.str());
         cerr << "ANALYZE TS 2" << endl;
      }
   
      /////////////////////////////////////////////////////////
      // Calculate averages of all channels		     //
      /////////////////////////////////////////////////////////
   
      // analyze all measurement channels
      if (!callChannelAnalysis (id, measnum, 
                           (channelAnalysis) &timeseries::sum)) {
         return false;
      }
   
      note = true;
      cerr << "ANALYZE TS 3" << endl;
      return true;
   }


   bool timeseries::sum (int resultnum, int measnum, 
                     string chnname, bool stim, const callbackarg& id)
   {
      cerr << "analyze " << chnname << " from " << measnum <<
         " into " << resultnum << endl;
   
      // get time series access class
      //const diagResult& aChn = diagChn::self();
      // get time series
      gdsDataObject* chndat = storage->findData (chnname);
      if (chndat == 0) {
         cerr << "analyze 1.5" << endl;
         return false;
      }
      // number of float values
      int N = tmps[resultnum].size;
      int num = tmps[resultnum].cmplx ? 2 * N : N;
   
      // get result access class
      const diagResult* aRes = 
         diagTimeSeries::self (stObjectTypeTimeSeries);
      if (aRes == 0) {
         cerr << "analyze 1.6" << endl;
         return false;
      }
   
      // get result object
      string		resname = 
         diagObjectName::makeName (stResult, rindex + resultnum);
      gdsDataObject* 	res = storage->findData (resname);
      cerr << "analyze 1" << endl;
   
      if (measnum == 0) {
         // init analysis result first time around
         if (res != 0) {
            storage->erase (resname);
         }
         int dim1 = includeStatistics ? 5 : N;
         int dim2 = includeStatistics ? N : 0;
         res = aRes->newObject (0, dim1, dim2, rindex + resultnum, -1,
                              tmps[resultnum].cmplx  ? gds_complex32: gds_float32);
         if (res != 0) {
            // set parameters of result
            string		name;
            int			i1, i2;
            if (!diagStorage::analyzeName (chnname, name, i1, i2)) {
               name = "";
            }
            aRes->clone (*res, chndat, false);
            res->setFlag (gdsDataObject::resultObj);
            aRes->setParam (*res, stTimeSeriesSubtype, 
                           tmps[resultnum].cmplx ? 1 : (includeStatistics ? 3 : 2));
            aRes->setParam (*res, stTimeSeriesAverageType, 
                           averageType);
            aRes->setParam (*res, stTimeSeriesAverages, averages);
            aRes->setParam (*res, stTimeSeriesf0, fZoom);
                           //1.0 / (measTime + mTimeAdd));
            aRes->setParam (*res, stTimeSeriesChannel, name);
            storage->addData (*res, false);
         }
      }
      if (res == 0) {
         cerr << "analyze 1.7" << endl;
         return false;
      }
      cerr << "analyze 2" << endl;
   
      // get time series
      memcpy (tmps[resultnum].x, chndat->value, num * sizeof (float));
   
      // apply filter if non-complex
      if (!tmps[resultnum].cmplx) {
         if (!tmps[resultnum].filter && 
            !tmps[resultnum].filterspec.empty() &&
            (tmps[resultnum].dx > 0)) {
            FilterDesign ds (1. / tmps[resultnum].dx);
            if (ds.filter (tmps[resultnum].filterspec.c_str())) {
               tmps[resultnum].filter = ds.release();
               cerr << "analyze 3: filter with " << filterSpec << endl;
            }
         }
         if (tmps[resultnum].filter) {
            try {
               tmps[resultnum].filter->reset();
               TSeries inp (Time (0), tmps[resultnum].dx, 
                           tmps[resultnum].size, tmps[resultnum].x);
               TSeries out = tmps[resultnum].filter->apply (inp);
               if ((int)out.getNSample() != tmps[resultnum].size) {
                  throw std::runtime_error("filter failed");
               }
               out.getData (tmps[resultnum].size, tmps[resultnum].x);
            }
               catch (...) {
                  cerr << "Filtering of time series failed" << endl;
               }
         }
      }
   
      // calculate mean and store result
      avg_specs		avgprm;
      avgprm.avg_type = (averageType != 1) ? 
         AVG_LINEAR_VECTOR : AVG_EXPON_VECTOR;
      avgprm.dataset_length = num;
      avgprm.data_type = DATA_REAL;
      avgprm.number_of_averages = averages;
      int num_so_far = measnum;
      if (avg (&avgprm, 1, tmps[resultnum].x, &num_so_far, 
              (float*) res->value) < 0) {
         return false;
      }
   
      // compute statistical properties
      if (!tmps[resultnum].cmplx && includeStatistics) {
      
         // minimum
         float* x = tmps[resultnum].x;
         float* y = (float*) res->value + 2 * num;
         for (int i = 0; i < num; i++) {
            if ((measnum == 0) || (x[i] < y[i])) {
               y[i] = x[i];
            }
         }
         // maximum
         y = (float*) res->value + 3 * num;
         for (int i = 0; i < num; i++) {
            if ((measnum == 0) || (x[i] > y[i])) {
               y[i] = x[i];
            }
         }
         // rms
         float* xsqr = tmps[resultnum].xsqr;
         y = (float*) res->value + 4 * num;
         for (int i = 0; i < num; i++) {
            x[i] *= x[i];
         }
         num_so_far = measnum;
         if (avg (&avgprm, 1, x, &num_so_far, xsqr) < 0) {
            return false;
         }
         for (int i = 0; i < num; i++) {
            y[i] = sqrt (xsqr[i]);
         }
      
         // std. dev.
         y = (float*) res->value + 1 * num;
         x = (float*) res->value;
         for (int i = 0; i < num; i++) {
            if (measnum == 0) {
               y[i] = 0;
            }
            else {
               y[i] = sqrt ((measnum + 1) / measnum * 
                           (xsqr[i] - x[i] * x[i]));
            }
         }
      }
      aRes->setParam (*res, stTimeSeriesAverages, 
                     (averageType != 1) ? measnum + 1 : 
                     min (measnum + 1, averages));
   
      cerr << "analyzed " << chnname << endl;
      return true;
   }


}
