/* version $Id: FilterFile.cc 6414 2011-04-01 18:11:04Z james.batch@LIGO.ORG $ */
/* -*- mode: c++; c-basic-offset: 3; -*- */
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <iostream>
#include <fstream>
#include "filterwiz/FilterFile.hh"
#include "iirutil.hh"
#include "memmap/mmap.hh"
#include <cstdlib>

   using namespace std;
   using namespace ligogui;

namespace filterwiz {


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Parsing utilities						        //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   typedef vector<string> tokenlist;

//______________________________________________________________________________
// Given a memory map of characters pointed to by p and ending at end, find the
// first instance of a '\n' or '\r'.  When that character is located, it will be
// the start of the next line (and also the return value of this function).
// Calculate the length of the line by searching for the next instance of a '\n'
// or '\r' character (or the end of the huge array), counting the characters until
// it's found.
   const char* nextline (const char* p, const char* end, 
                     int& linelen, bool skipleadingspace = false)
   {
      bool newline = false;
      for (; p < end; ++p) 
      {
         if ((*p == '\n') || (*p == '\r')) 
	 {
            newline = true;
         }
         else if (newline) 
	 {
            // Once newline is true, p points to the first character
            // beyond the newline.  If no leading whitespace is wanted,
            // keep advancing the pointer p until a non-whitespace character
            // is found.
            if (skipleadingspace) 
	    {
               for (; (p < end) && isspace (*p); ++p) ;
            }
            // Find the next newline or end of file, counting the characters
            // so we know how long this line is.
            for (linelen = 0; (p + linelen < end) &&
                (p[linelen] != '\n') && (p[linelen] != '\r'); ++linelen) ;
            return p;
         }
      }
      return end;	// What is linelen a this point?
   }

//______________________________________________________________________________
// p is a character array that has been allocated in FilterFile::write(filename).
// The size of the array starts at 131072.  end points to the last character in the
// array, and line is the text that needs to be added to the array.  If the line to
// be added to p is longer than the space remaining (end-p), completely fill the
// array, but don't copy any more!  Successive calls to putline will not put any
// more characters in the array under this condition.
   char* putline (char* p, const char* end, const char* line)
   {
      int len = strlen (line);
      if (len >= end - p) {
         memcpy (p, line, end - p);
         return p + (end - p);
      }
      else {
         memcpy (p, line, len);
         p[len] = '\n';
         return p + len + 1;
      }
   }

//______________________________________________________________________________
// Create a set of strings from a line of characters where each string is
// a set of characters separated by whitespace. 
   void tokens (const char* p, int max, tokenlist& tok, bool add = false)
   {
      if (!add) tok.clear();
      const char* end = p + max;
      const char* first = 0;
      for (; p < end; ++p) {
         if (isspace (*p)) {
            if (first) {
               tok.push_back (string (first, p - first));
               first = 0;
            }
         }
         else if (!first) {
            first = p;
         }
      }
      if (first) {
         tok.push_back (string (first, p - first));
      }
   }

//______________________________________________________________________________
// bool isname(const char *p) takes a pointer to a conventional C string and
// evaluates it to see if it's a proper name.  Names start with an alpha and
// are followed by 0 or more alpha, numeral, or '_' characters.
   bool isname (const char* p) 
   {
      if (!isalpha (*p)) {
         return false;
      }
      for (; *p; ++p) {
         if (!isalnum (*p) && !(*p == '_')) {
            return false;
         }
      }
      return true;
   }

//______________________________________________________________________________
   bool isname  (const string& p) 
   {
      return isname (p.c_str());
   }

//______________________________________________________________________________
// bool isintnum(const char *p) checks to see if the string pointed to by p
// represents an integer number.  An integer number is at least 1 digit
// '0' through '9', optionally preceded by a '+' or '-' character.
   bool isintnum  (const char* p) 
   {
      int where = 0; // 0 - start; 1 - after sign, 2 - after 1st digit
      for (; *p; ++p) {
         if (isdigit (*p)) {
            if (where <= 1) where = 2;
         }
         else if ((*p == '-') || (*p == '+')) {
            if (where == 0) where = 1;
            else 
               return false;
         }
         else {
            return false;
         }
      }
      return where == 2;
   }

//______________________________________________________________________________
   bool isintnum  (const string& p) 
   {
      return isintnum (p.c_str());
   }

//______________________________________________________________________________
   bool isnum  (const char* p) 
   {
      int where = 0; // 0 - start; 1 - after sign, 2 - after 1st digit
      // 3 - after point; 4 - after fraction didit
      // 5 - after expo, 6 - after expo sign, 7 - after expo digit
      for (; *p; ++p) {
         if (isdigit (*p)) {
            if (where <= 1) where = 2;
            else if (where == 3) where = 4;
            else if ((where == 5) || (where == 6)) where = 7;
         }
         else if ((*p == '-') || (*p == '+')) {
            if (where == 0) where = 1;
            else if (where == 5) where = 6;
            else 
               return false;
         }
         else if (*p == '.') {
            if (where == 2) where = 3;
            else 
               return false;
         }
         else if ((*p == 'e') || (*p == 'E')) {
            if ((where >= 2) && (where <= 4)) where = 5;
            else 
               return false;
         }
         else {
            return false;
         }
      }
      return (where == 2) || (where == 3) || (where == 4) || 
         (where == 7);
   }

//______________________________________________________________________________
   bool isnum  (const string& p) 
   {
      return isnum (p.c_str());
   }

//______________________________________________________________________________
// bool ismodname(const char *p) checks to see if the string pointed to by p
// is a proper module name.  This follows the same rule as above for a name,
// which is an alpha optionally followed by one or more alpha, numeral, or '_'
// characters.
   bool ismodname (const char* p) 
   {
      return isname (p);
   }

//______________________________________________________________________________
   bool ismodname  (const string& p) 
   {
      return ismodname (p.c_str());
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// FilterFile							        //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
// A FilterFile has a couple of protected members, fFilename which is a string,
// and fModules which is a FilterModuleList.
   FilterFile::FilterFile ()
   {
      legacy_write = 0 ; // JCB
      gain_only_allowed = 1 ; // JCB
   }

//______________________________________________________________________________
   void FilterFile::clear()
   {
      fModules.clear();
   }

//______________________________________________________________________________
// Set the sample rate for the file given a character string
   void FilterFile::setFSample(const char *sample)
   {
      if (sample)
	 fSample = sample ;
      else
	 fSample = "" ;
   }

// Set the sample rate for the file given an unsigned int
   void FilterFile::setFSample(unsigned int sample)
   {
      char buf[16] ;

      sprintf(buf, "%u", sample) ;
      setFSample((const char *) buf) ;
   }

// Set the sample rate for the file given a double
   void FilterFile::setFSample(double sample)
   {
      setFSample((unsigned int) sample) ;
   }

//______________________________________________________________________________
// Either replace an existing FilterModule of the same name if it exists, or
// add a new FilterModule to the vector of FilterModules fModules.
   void FilterFile::add (const char* name, double fsample)
   {
      // Search the vector of FilterModule instances in fModules to see if the
      // name exists.  If it does, the FilterModule in the vector is replaced
      // with a new FilterModule of the same name and the specified fsample.
      for (FilterModuleList::iterator i = fModules.begin(); 
          i != fModules.end(); ++i) {
         if (strcmp (i->getName(), name) == 0) {
            *i = FilterModule (name, fsample);
            return;
         }
      }
      // The name wasn't found, add the new FilterModule to the end of the vector
      // of FilterModules, fModules.
      fModules.push_back (FilterModule (name, fsample));
   }

//______________________________________________________________________________
   void FilterFile::remove (const char* name)
   {
      for (FilterModuleList::iterator i = fModules.begin(); 
          i != fModules.end(); ++i) {
         if (strcmp (i->getName(), name) == 0) {
            fModules.erase (i);
            return;
         }
      }
   }

//______________________________________________________________________________
   FilterModule* FilterFile::find (const char* name)
   {
      for (FilterModuleList::iterator i = fModules.begin(); 
          i != fModules.end(); ++i) {
         if (strcmp (i->getName(), name) == 0) {
            return &*i;
         }
      }
      return 0;
   }

//______________________________________________________________________________
   const FilterModule* FilterFile::find (const char* name) const
   {
      for (FilterModuleList::const_iterator i = fModules.begin(); 
          i != fModules.end(); ++i) {
         if (strcmp (i->getName(), name) == 0) {
            return &*i;
         }
      }
      return 0;
   }

//______________________________________________________________________________
   bool FilterFile::valid (std::string& errmsg) const
   {
      errmsg = "";
      // loop over modules/sections
      for (FilterModuleList::const_iterator mod = fModules.begin();
          mod != fModules.end(); ++mod) {
         for (int i = 0; i < kMaxFilterSections; ++i) {
            if (!(*mod)[i].valid()) {
               if (!errmsg.empty()) errmsg += "\n";
               errmsg += mod->getName();
               char buf [256];
               sprintf (buf, "/%i", i);
               errmsg += buf; 
            }
         }
      }
      return errmsg.empty();
   }

//______________________________________________________________________________
// Update each filter section in each module in the file.
   bool FilterFile::update()
   {
      // loop over modules/sections
      for (FilterModuleList::iterator mod = fModules.begin();
          mod != fModules.end(); ++mod) {
         for (int i = 0; i < kMaxFilterSections; ++i) {
	    // mod is a FilterModule, (*mod)[i] is a FilterSection
            if (!(*mod)[i].update()) {
               return false; 
            }
         }
      }
      return true;
   }

//______________________________________________________________________________
   bool FilterFile::read (const char* filename)
   {
      cerr << "read " << filename << endl ;
      // Note that the file name is the full path of the file.
      if (!filename) {
         cerr << "Illegal file name" << endl;
         return false;
      }
      string oldfilename = fFilename;
      fFilename = filename;
      gdsbase::mmap mfile (filename, std::ios_base::in);
      if (!mfile) {
         cerr << "Unable to open file " << filename << endl;
         fFilename = oldfilename;
         return false;
      }
      // JCB - start
      if (stat(fFilename.c_str(), &fStat))
      {
	 // An error occurred getting stat.
	 cerr << "stat error, " << strerror(errno) << endl ;
      }
      // JCB - end
      bool succ = read ((const char*)mfile.get(), mfile.size()) == 
         (int)mfile.size();
      if (!succ) {
         fFilename = oldfilename;
      }
      return succ;
   }

//______________________________________________________________________________
   int FilterFile::read (const char* beg, int maxlen)
   {
      // clear old modules
      fModules.clear();
      // check magic string
      if ((maxlen < (int)strlen (kMagic)) ||
         (strncmp (beg, kMagic, strlen (kMagic)) != 0)) {
         clear();
         cerr << "Not an online filter file" << endl;
         return 0;
      }
      // first look for filter module names
      const char* end = beg + maxlen;
      int linelen = 0;
      int lenModDef = strlen(kModuleDef);  //kModuleDef is "# MODULES" from FilterModule.hh
      for (const char* p = beg; p < end; p = nextline (p, end, linelen)) {
         if (linelen > lenModDef && strncmp (p, kModuleDef, lenModDef) == 0) {
            tokenlist tok;
            p += lenModDef;
            linelen -= lenModDef;
            tokens (p, linelen, tok);
            for (tokenlist::iterator i = tok.begin(); 
                i != tok.end(); ++i) {
               if (ismodname (i->c_str())) {
                  add (i->c_str(), 1E-12);
               }
               else {
                  cerr << "Illegal file module name " << *i << endl;
               }
            }
         }
      }
   
      /* JCB for aLIGO, specify one sample rate for the entire file.  Look for
       * any line that starts with kModuleSamplingDef and is followed either by a 
       * number or some token followed by a number.  That should handle both legacy
       * files (where a sampling rate is specified for each module) and aLIGO
       * files where only one sample rate is specified near the beginning of the file.
       * (actually, it could be anywhere in the file, but by convention put it near the top
       * so we can find it.)
       */
      if (!legacy_write)
      {
	 double	sampleRate = 2048 ; // Default value

	 // Look anywhere in the file for a sampling rate 
	 for (const char *p = beg; p < end; p = nextline (p, end, linelen))
	 {
	    // kModuleSamplingDef is "# SAMPLING" from FilterModule.hh
	    if ((linelen > (int) strlen(kModuleSamplingDef)) &&
	       (strncmp (p, kModuleSamplingDef, strlen(kModuleSamplingDef)) == 0))
	    {
	       tokenlist tok ;
	       tokens (p + strlen(kModuleSamplingDef), linelen - strlen(kModuleSamplingDef), tok) ;
	       if ((tok.size() == 1) && isnum (tok[0]))
	       {
		  sampleRate = atof(tok[0].c_str()) ;
		  break ;
	       }
	       else if ((tok.size() == 2) && isnum (tok[1]))
	       {
		  // We really don't care what tok[0] was.
		  sampleRate = atof(tok[1].c_str()) ;
		  break ;
	       }
	       else
	       {
		  cerr << "Illegal sampling rate specification:" << endl 
		       << "  " << string (p, linelen) << endl ;
	       } 
	    }
	 }
	 // Assign the sample rate to the file.
	 setFSample(sampleRate) ;

	 // Now assign the sample rate to all the modules for legacy.
	 for (FilterModuleList::iterator i = modules().begin(); 
	       i != modules().end(); i++)
	 {
	    i->setFSample(sampleRate) ;
	 }
      }
      else
      {
	 // next look for filter module sampling rate
	 for (const char* p = beg; p < end; p = nextline (p, end, linelen)) {
	    // kModuleSamplingDef is "# SAMPLING" from FilterModule.hh
	    if ((linelen > (int)strlen (kModuleSamplingDef)) &&
	       (strncmp (p, kModuleSamplingDef, 
			strlen (kModuleSamplingDef)) == 0)) {
	       tokenlist tok;
	       tokens (p + strlen (kModuleSamplingDef), 
		      linelen - strlen (kModuleSamplingDef), tok);
	       if ((tok.size() == 2) && ismodname (tok[0]) && isnum (tok[1])) {
		  FilterModule* mod = find (tok[0].c_str());
		  if (mod) {
		     mod->setFSample (atof (tok[1].c_str()));
		     // Also set the sample rate for the file.
		     setFSample(tok[1].c_str()) ;
		  }
		  else {
		     cerr << "Sampling rate for unknown filter module " 
			<< tok[0] << endl;
		  }
	       }
	       else {
		  cerr << "Illegal sampling rate specification:" << endl 
		     << "  " << string (p, linelen) << endl;
	       }
	    }
	 }
      }
   
      // if we got no filter modules, start a scan
      if (fModules.empty()) {
         for (const char* p = beg; p < end; p = nextline (p, end, linelen)) {
            if ((*p == '#') || isspace (*p)) {
               continue;
            }
            tokenlist tok;
            tokens (p, linelen, tok);
            if ((tok.size() > 0) && ismodname (tok[0].c_str())
               && !find (tok[0].c_str())) {
               cout << "Warning: Filter module " << tok[0] << 
                  " added." << endl;
               add (tok[0].c_str(), 1E-12);
            }
         }
      }
   
      // need at least one module
      if (fModules.empty()) {
         clear();
         cerr << "No filter module specification" << endl;
         return 0;
      }
   
      // guess sample rate if necessary
      for (FilterModuleList::iterator i = modules().begin(); 
          i != modules().end(); ++i) {
         if (i->getFSample() <= 1E-10) {
            i->guessFSample (fFilename.c_str());
            cout << "Warning: Guessing sampling rate of " << i->getName() <<
               " to be " << i->getFSample() << endl;
         }
      }
   
      // now read modules/sections
      string header;
      for (const char* p = nextline (beg, end, linelen); 
          p < end; p = nextline (p, end, linelen)) {
         // skip filter info lines
	 // kModuleDef is "# MODULES"
	 // kModuleSamplingDef is "# SAMPLING"
	 // kSectionDesignDef is "# DESIGN" from FilterModules.hh
         if (((linelen > (int)strlen (kModuleDef)) && 
             (strncmp (p, kModuleDef, strlen (kModuleDef)) == 0)) ||
            ((linelen > (int)strlen (kModuleSamplingDef)) &&
            (strncmp (p, kModuleSamplingDef, 
                     strlen (kModuleSamplingDef)) == 0)) ||
            ((linelen > (int)strlen (kSectionDesignDef)) &&
            (strncmp (p, kSectionDesignDef, 
                     strlen (kSectionDesignDef)) == 0))) {
            continue;
         }
         // add comments to header
         if (*p == '#') {
            if (header.empty()) header += '\n';
            header += string (p, linelen);
            continue;
         }
         // found something
         tokenlist tok;
         tokens (p, linelen, tok);
         if (tok.empty()) {
            header += '\n';
            continue;
         }
         // check if correctly formated section spec
	 // This assumes every filter section will have
	 // at least one SOS, since the last 4 tokens
	 // are the b1, b2, a1, a2 coefficients and
	 // the tok.size() needs to be 12.
         if ((tok.size() != 12) || !ismodname (tok[0]) || 
            !isintnum (tok[1]) || !isintnum (tok[2]) ||
            !isintnum (tok[3]) || !isnum (tok[4]) ||
            !isnum (tok[5]) || !isnum (tok[7]) ||
            !isnum (tok[8]) || !isnum (tok[9]) ||
            !isnum (tok[10]) || !isnum (tok[11])) {
            cerr << "Illegal filter section:" << endl <<
               "  " << string (p, linelen) << endl;
            continue;
         }
         FilterModule* mod = find (tok[0].c_str());
         if (!mod) {
            cerr << "Unknown filter module " << tok[0] << endl;
            continue;
         }
         int index = atoi (tok[1].c_str());
         if ((index < 0) || (index >= kMaxFilterSections)) {
            cerr << "Illegal section number " << tok[1] << endl;
            continue;
         }
         input_switching  inpsw =  
            (input_switching)(atoi (tok[2].c_str()) / 10);
         if ((inpsw != kAlwaysOn) && (inpsw != kZeroHistory)) {
            cerr << "Illegal input switch " << tok[2] << endl;
            continue;
         }
         output_switching outsw = 
            (output_switching)(atoi (tok[2].c_str()) % 10);
         if ((outsw != kImmediately) && (outsw != kRamp) &&
            (outsw != kInputCrossing) && (outsw != kZeroCrossing)) {
            cerr << "Illegal output switch " << tok[2] << endl;
            continue;
         }
         int sosnum = atoi (tok[3].c_str());
//         if ((sosnum < 0) || (sosnum > kMaxFilterSOS)) {
// Change minimum sosnum to 1, front-ends can't load 0 SOS.	 
	 if (sosnum == 0)
	 {
	    sosnum = 1 ; // Leap of faith...  we hope there are coefficients to 
	    		 // go with this.  Gain only might have been written with
			 // 0 previously, this will attempt to repair the file.
	 }
         if ((sosnum < 1) || (sosnum > kMaxFilterSOS)) {
            cerr << "Illegal SOS number " << tok[3] << endl;
            continue;
         }
         double ramp = 0;
         double tolerance = 0;
         double timeout = 0;
         switch (outsw) {
            case kImmediately: 
               break;
            case kRamp:
               ramp = atof (tok[4].c_str()) / mod->getFSample();
               if (ramp <= 0) {
                  cerr << "Illegal ramp time " << tok[4] << endl;
                  continue;
               }
               break;
            case kInputCrossing:
            case kZeroCrossing:
               tolerance = atof (tok[4].c_str());
               timeout = atof (tok[5].c_str()) / mod->getFSample();
               if (tolerance < 0) {
                  cerr << "Illegal switch tolerance " << tok[4] << endl;
                  continue;
               }
               if (timeout < 0) {
                  cerr << "Illegal timeout " << tok[5] << endl;
                  continue;
               }
               break;
         }
// Before continuing, check for gain-only filter.  This is a special case
// of sosnum == 1 and tok[8], tok[9], tok[10], tok[11] all equal 0.0.
	 if (sosnum == 1 && atof(tok[8].c_str()) == 0.0 && atof(tok[9].c_str()) == 0.0
			 && atof(tok[10].c_str()) == 0.0 && atof(tok[11].c_str()) == 0.0)
	 {
	    // Gain-only filter.  Change the sosnum to 0 to keep this
	    // from being mis-interpreted internally.
	    sosnum = 0;
	 }
         // OK, we got the 1st line: Now check for additional SOS lines
         for (int i = 1; i < sosnum; ++i) {
            p = nextline (p, end, linelen);
            if (p >= end) {
               cerr << "Not enough lines in file" << endl;
               break;
            }
            tokens (p, linelen, tok, true);
         }
         if ((sosnum > 0) && ((int)tok.size() != 8 + 4 * sosnum)) {
            cerr << "Not enough second order sections for " 
               << tok[0] << endl;
            continue;
         }
         bool ok = true;
         for (int i = 12; i < 8 + 4 * sosnum; ++i) {
            if (!isnum (tok[i])) ok = false;
         }
         if (!ok) {
            cerr << "Filter coefficients for " << tok[0] 
               << " are not all numbers" << endl;
            continue;
         }
         // Finally we have a valid filter section!
         FilterSection& sect = mod->operator[] (index);
         sect.filter().reset();
         sect.filter().gain (atof (tok[7].c_str()));
         for (int i = 8; i < 8 + 4 * sosnum; i += 4) {
            sect.filter().biquad (1., atof (tok[i+2].c_str()),
                                 atof (tok[i+3].c_str()),
                                 atof (tok[i].c_str()),
                                 atof (tok[i+1].c_str()));
         }
         sect.setName (tok[6].c_str());
         sect.setInputSwitch (inpsw);
         sect.setOutputSwitch (outsw);
         sect.setRamp (ramp);
         sect.setTolerance (tolerance);
         sect.setTimeout (timeout);
         sect.setHeader (header.c_str());
         header = "";
      }
   
      // last look for filter section design strings
      for (const char* p = beg; p < end; p = nextline (p, end, linelen)) {
         // look for design string
         if ((linelen > (int)strlen (kSectionDesignDef)) &&
            (strncmp (p, kSectionDesignDef, 
                     strlen (kSectionDesignDef)) == 0)) {
            // tokenize string
            tokenlist tok;
            tokens (p + strlen (kSectionDesignDef), 
                   linelen - strlen (kSectionDesignDef), tok);
            if ((tok.size() >= 3) && ismodname (tok[0]) &&
               isintnum (tok[1])) {
               // found a valid design
               FilterModule* mod = find (tok[0].c_str());
               int index = atoi (tok[1].c_str());
               if (mod && (index >= 0) && (index < kMaxFilterSections)) {
                  // found a valid module
                  string s (p + strlen (kSectionDesignDef),
                           linelen - strlen (kSectionDesignDef));
                  int indent = strlen (kSectionDesignDef);
                  // delete module name
                  string::size_type pos = s.find (tok[0].c_str());
                  if (pos != string::npos) {
                     s.erase (0, pos + tok[0].size());
                     indent += pos + tok[0].size();
                  }
                  // delete section number
                  pos = s.find (tok[1].c_str());
                  if (pos != string::npos) {
                     s.erase (0, pos + tok[1].size());
                     indent += pos + tok[1].size();
                  }
                  // delete leading white space
                  while (!s.empty() && isspace (s[0])) {
                     s.erase (0, 1);
                     ++indent;
                  }
                  // delete trailing white space
                  while (!s.empty() && isspace (s[s.size()-1])) 
                     s.erase (s.size()-1, 1);
                  // look for line continuation
                  while (!s.empty() && (s[s.size()-1] == '\\') && 
                        (p < end)) {
                     // remove continuation characters
                     do {
                        s.erase (s.size()-1, 1);
                     } while (!s.empty() && (isspace (s[s.size()-1])));
                     // add new line
                     s += "\n";
                     // look for new line
                     p = nextline (p, end, linelen);
                     string n (p, linelen);
                     // delete leading white space and #
                     for (int i = 0; (i < indent) && !n.empty() && 
                         (isspace (n[0]) || (n[0] == '#')); ++i) {
                        n.erase (0, 1);
                     }
                     // delete trailing white space
                     while (!n.empty() && isspace (n[n.size()-1])) 
                        n.erase (n.size()-1, 1);
                     s += n;
                  }
                  // set design string
                  (*mod)[index].setDesign (s.c_str());
               }
               else if (!mod) {
                  cerr << "Design for unknown filter module " 
                     << tok[0] << endl;
               }
               else {
                  cerr << "Illegal section number " << tok[0] <<
                     "/" << tok[1] << endl;
               }
            }
            else {
               cerr << "Illegal design specification:" << endl 
                  << "  " << string (p, linelen) << endl;
            }
         }
      }
   
      // now check filter if design strings are ok
      for (FilterModuleList::iterator i = fModules.begin();
          i != fModules.end(); ++i) {
         i->checkDesign();
      }
   
      return maxlen;
   }

//______________________________________________________________________________
   bool FilterFile::write (const char* filename)
   {
      cerr << "write " << filename << endl ;
      if (!filename) {
         cerr << "Illegal file name" << endl;
         return false;
      }
      if (!update()) {
         cerr << "Invalid filter design" << endl;
         return false;
      }
      ofstream out (filename);
      if (!out) {
         cerr << "Unable to open file " << filename << endl;
         return false;
      }
      int size = 64*1024;
      char* p  = 0;
      int len;
      do {
         if (p) delete [] p;
         size *= 2;
         p = new char [size];
         if (!p)
         {
                cerr << "Memory exhausted attempting to write file" << filename << endl ;
                return (false) ;
	 }
         // Assuming we got the memory we requested, attempt to fill the array p with
         // the text that will be a filter file. If we need to write more characters
         // than were allocated, delete the whole thing, allocate twice as much memory,
         // and try again.

         len = write (p, size);
      } while (len >= size);
      out.write (p, len);
      delete [] p;
      if (!out) {
         cerr << "Unable to write file " << filename << endl;
      }
      return (out);
   }

//______________________________________________________________________________
   int FilterFile::write (char* beg, int maxlen)
   {
      // Format of filter definitions and coefficients.  These are extended precision. - JCB
      const char *format_str1 = "%-8s %i %2i %i %6i %6i %-10s %22.16g %20.16f %20.16f %20.16f %20.16f" ;
      const char *format_str2 = "%63s %20.16f %20.16f %20.16f %20.16f" ;

      if (!update()) {
         cerr << "Invalid filter design" << endl;
         return false;
      }

      // Decide what output format to use. - JCB
      if (legacy_write)
      {
	 // Replace the format with the original strings.
	 format_str1 = "%-8s %i %2i %i %6i %6i %-10s %22.14g %20.14f %20.14f %20.14f %20.14f" ;
	 format_str2 = "%63s %20.14f %20.14f %20.14f %20.14f" ;
      }
   
      char* p = beg;
      char* end = beg + maxlen;
      // write magic header
      p = putline (p, end, kMagic);
      // write warning
      p = putline (p, end, 
                  "#\n# Computer generated file: DO NOT EDIT\n#");
      // write filter module names
      string line = kModuleDef;
      for (FilterModuleList::const_iterator mod = fModules.begin();
          mod != fModules.end(); ++mod) {
         line += string (" ") + mod->getName();
         if (line.size() > 70) {
            p = putline (p, end, line.c_str());
            line = kModuleDef;
         }
      }
      if (line.size() > strlen (kModuleDef)) {
         p = putline (p, end, line.c_str());
      }
      p = putline (p, end, "#");
      /* JCB For aLIGO write the sample rate once for the file since
       * all modules use the same sample rate.
       */
      if (!legacy_write)
      {
	 char buf[256] ;
	 // Write the sampling rate/
	 sprintf(buf, "%s RATE %s", kModuleSamplingDef, getFSampleStr()) ;
	 p = putline(p, end, buf) ;
	 p = putline(p, end, "#") ;
      }
      // write filter sections
      for (FilterModuleList::const_iterator mod = fModules.begin();
          mod != fModules.end(); ++mod) {
      
         // write title
         char buf[256];
         p = putline (p, end, "##########################################"
                     "######################################");
         sprintf (buf, "### %-72s ###", mod->getName());
         p = putline (p, end, buf);
         p = putline (p, end, "##########################################"
                     "######################################");
      
         // write sampling rate
	 // JCB For aLIGO, don't write the sample rate.  It's written
	 // once for the file after the module names.
	 if (legacy_write)
	 {
	    sprintf (buf, "%s %s %g", kModuleSamplingDef,
		    mod->getName(), mod->getFSample());
	    p = putline (p, end, buf);
	 }
      
         // write design string
         for (int i = 0; i < kMaxFilterSections; ++i) {
            const FilterSection& sect = mod->operator[] (i);
            if (!sect.empty()) {
               string s;
               char buf[256];
               sprintf (buf, "%s   %s %i ", kSectionDesignDef,
                       mod->getName(), i);
               s += buf;
               string linebreak = " \\## ";
               linebreak.append (s.size() - 2, ' ');
               string cmd = sect.getDesign();
               string::size_type pos;
               while ((pos = cmd.find ('\n')) != string::npos) {
                  cmd.erase (pos, 1);
                  cmd.insert (pos, linebreak.c_str());
               }
               while ((pos = cmd.find ("##")) != string::npos) {
                  cmd[pos] = '\n';
               }
               s += cmd;
               p = putline (p, end, s.c_str());
            }
         }
         sprintf (buf, "### %-72s ###", "");
         p = putline (p, end, buf);
      
         // write coefficients
         for (int i = 0; i < kMaxFilterSections; ++i) {
            const FilterSection& sect = mod->operator[] (i);
            if (!sect.empty()) {
	       //--- iirorder suppresses zeroes at fNy ... we don't want this
               //int order = iirorder (sect.filter().get());
               int order = iirsoscount(sect.filter().get());
               if (order < 0) {
                  cerr << "Not an IIR filter" << endl;
                  continue;
               } 
	       //--- Reserve space for zeroes.
	       else if (!order) {
		  order = 1;
	       }
               int nba;
               double* coeff = new double [1 + 4*order];
	       coeff[0] = 1.0;

               if (!iir2z (sect.filter().get(), nba, coeff, "o")) {
                  cerr << "Unable to obtain online filter coefficients" 
		       << endl;
		  delete[] coeff;
                  continue;
               }
               //p = putline (p, end, sect.getHeader());
               char buf[1024];
               int sosnum = (nba - 1) / 4;
               if (sosnum == 0) {
// aLIGO front ends can't handle 0 SOS.  So even if there aren't any SOS
// and we want a gain only filter, we have to set SOS to 1 with coefficients
// to 0.0.
//                  sosnum = 0;
                  sosnum = 1 ;
                  coeff[1] = coeff[2] = coeff[3] = coeff[4] = 0.0;
               }
	       // Added test for legacy write and sosnum < 1  - JCB
	       // sosnum should never be less than 0.
	       // If legacy_write is set, sosnum should never be less than 1.
	       // If not legacy_write && gain_only isn't allowed, sosnum should not be less than 1
               if ((sosnum < 0) || 
		   (legacy_write && sosnum < 1) || 
		   (!legacy_write && !gain_only_allowed && (sosnum < 1)) || 
		   (sosnum > kMaxFilterSOS)) 
	       {
                  cerr << "Invalid number of SOSs " << sosnum << endl;
		  delete[] coeff;
                  continue;   // Don't write the filter section.
               }
               int arg1 = 0;
               int arg2 = 0;
               int sw = 10*int(sect.getInputSwitch()) +
                  int(sect.getOutputSwitch());
               switch (sect.getOutputSwitch()) {
                  case kImmediately: 
                     break;
                  case kRamp:
                     arg1 = int(sect.getRamp() * mod->getFSample() + 0.5);
                     break;
                  case kInputCrossing:
                  case kZeroCrossing:
                     arg1 = (int)(sect.getTolerance() + 0.5);
                     arg2 = (int)(sect.getTimeout() * mod->getFSample()  + 0.5);
                     break;
               }
               string sect_name = validSectionName (sect.getName());
	       // Changed to use a format string instead of a fixed format. - JCB
               sprintf (buf, format_str1, 
                       mod->getName(), i, sw, sosnum, arg1, arg2, 
                       sect_name.c_str(), coeff[0], coeff[1], coeff[2],
                       coeff[3], coeff[4]);
               p = putline (p, end, buf);
               for (int j = 1; j < sosnum; ++j) {
                  sprintf (buf, format_str2,
                          "", coeff[1+4*j], coeff[2+4*j], 
                          coeff[3+4*j], coeff[4+4*j]);
                  p = putline (p, end, buf);
               }
	       delete[] coeff;
            }
         }
      
         // empty line
         p = putline (p, end, "");
      }
   
      return p - beg;
   }

//______________________________________________________________________________
   std::string FilterFile::validSectionName (const char* name)
   {
      std::string s;
      for (const char* p = name; *p != 0; ++p) {
         if (isgraph (*p)) {
            s += *p;
         }
      }
      if (s.empty()) s = "Unknown";
      return s;
   }

// JCB - start
//______________________________________________________________________________
   bool FilterFile::checkFileStat()
   {
      struct stat buf ;

      if (stat(fFilename.c_str(), &buf))
      {
	 // An error occurred getting stat.
	 cerr << "stat error for  " << fFilename << ": " << strerror(errno) << endl ;
      }
      // Compare the st_ino and st_mtime fields.  They need to be the same.
      cerr << (buf.st_ino != fStat.st_ino ? "file st_ino differs" : "file st_ino ok") << endl ;
      cerr << (buf.st_mtime != fStat.st_mtime ? "file st_mtime differs" : "file st_mtime ok") << endl ;
      return (buf.st_ino == fStat.st_ino && buf.st_mtime == fStat.st_mtime) ;
   }

//______________________________________________________________________________
   // This needs to be called after a file is saved since the inode number and mtime
   // for the file will have changed.
   void FilterFile::updateFileStat(const char *path)
   {
      if (path)
      {
	 if (stat(path, &fStat))
	 {
	    // An error occurred getting stat.
	    cerr << "stat error for  " << path << ": " << strerror(errno) << endl ;
	 }
      }
      return ;
   }

// JCB - end


//______________________________________________________________________________
    // JCB Implementation of SectCopy class.  A vector of these classes is created
    // when the Copy button is pressed in the Sections frame.  This should capture
    // enough data to allow a section to be created in another module (or section)
    // by pressing the paste button.  If a single section is copied, it can be pasted
    // in any filter section, the fIndex value would be ignored.  If multiple sections
    // are copied, a Paste operation will create the sections using the fIndex value
    // to put them in the same place in whatever module they are pasted in.

    // Constructor takes a pointer to a filter section.
    SectCopy::SectCopy(FilterSection *section)
    {
        if (section)
        {
            fIndex = section->getIndex() ;
            fInp_sw = section->getInputSwitch() ;
            fOut_sw = section->getOutputSwitch() ;
            fRamp = section->getRamp() ;
            fTolerance = section->getTolerance() ;
            fTimeout = section->getTimeout() ;
            fName = section->getName() ;
            fDesign = section->getDesign() ;
        }
        else
	    // This shouldn't happen.
            cerr << "SectCopy() - section pointer is null!" << endl ;
    }

    // Set the index explicitly.  Allowable values are -1, 0-9.
    void SectCopy::SetIndex(int index)
    {
        // Make sure the value passed is legal.
        if (index < -1 || index > 9)
            index= -1 ;
        fIndex = index ;
        return ;
    }

    // Copy the saved design information to the filter section.
    void SectCopy::PasteSection(FilterSection *section)
    {
        if (section)
        {
            section->setInputSwitch(fInp_sw) ;
            section->setOutputSwitch(fOut_sw) ;
            section->setRamp(fRamp) ;
            section->setTolerance(fTolerance) ;
            section->setTimeout(fTimeout) ;
            section->setName(fName.c_str()) ;
            section->setDesign(fDesign.c_str()) ;
        }
        return ;
    }

    // This is the list of copied filter sections.  It's a vector.
    SectCopyList fSectCopyList ;
// JCB - end

} // namespace filterwiz
