/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: TLTFPainter						*/
/*                                                         		*/
/* Module Description: Time frequency histogram painter.		*/
/*                                                         		*/
/* Revision History:					   		*/
/* Rel   Date     Programmer  	Comments				*/
/* 0.1	 3Apr01   D. Sigg    	First release		   		*/
/*                                                         		*/
/* Documentation References:						*/
/*	Man Pages: TLTFPainter.html					*/
/*	References: none						*/
/*                                                         		*/
/* Author Information:							*/
/* Name          Telephone       Fax             e-mail 		*/
/* Daniel Sigg   (509) 372-8132  (509) 372-8137  sigg_d@ligo.mit.edu	*/
/*                                                         		*/
/*                                                         		*/
/*                      -------------------                             */
/*                                                         		*/
/*                             LIGO					*/
/*                                                         		*/
/*        THE LASER INTERFEROMETER GRAVITATIONAL WAVE OBSERVATORY.	*/
/*                                                         		*/
/*                     (C) The LIGO Project, 1999.			*/
/*                                                         		*/
/*                                                         		*/
/* Caltech				MIT		   		*/
/* LIGO Project MS 51-33		LIGO Project NW-17 161		*/
/* Pasadena CA 91125			Cambridge MA 01239 		*/
/*                                                         		*/
/* LIGO Hanford Observatory		LIGO Livingston Observatory	*/
/* P.O. Box 1970 S9-02			19100 LIGO Lane Rd.		*/
/* Richland WA 99352			Livingston, LA 70754		*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#ifndef _LIGO_TLTFPainter
#define _LIGO_TLTFPainter

#include "TVirtualPad.h"
#include "GuiTypes.h"
#include "THistPainter.h"
#include "TH2.h"


/** Paints and scrolls a histogram.
    @memo Helper class to draw time frequency plots.
    @author Written April 2001 by Daniel Sigg
    @version 1.0
 ************************************************************************/

   class TLTFPainter : public THistPainter {
   public:
      /// Default constructor
      TLTFPainter();
      /// Destructor
      virtual ~TLTFPainter();
      /// Overrides table init and allows partial repaint
      virtual Int_t TableInit();
      /// Update the horizontal scroll vallue
      virtual void HScroll (Int_t bins);
      /// Executes an events (handles scroll events)
      virtual void ExecuteEvent (Int_t event, Int_t px, Int_t py);
   
   protected:
      /// Do we need to scroll horizontally
      Bool_t		fHScroll;
      /// Horizontal scroll value
      Int_t		fHScrollNum;
   
    //ClassDef(TLTFPainter,0)  //helper class to draw histograms
   };


/** Time frequency plot types.
    @memo Plot types for time frequency histograms.
    @author Written April 2001 by Daniel Sigg
    @version 1.0
 ************************************************************************/
   enum tftype_t {
   /// Color coded time frequency plot
   kTfColor = 0,
   /// Scatter plot like time frequency plot
   kTfScatter = 1,
   /// Box like time frequency plot
   kTfBox = 2,
   /// Color coded time frequency plot with z scale
   kTfColorZ = 3
   };


/** A time frequency histogram.
    @memo Time frequency class.
    @author Written April 2001 by Daniel Sigg
    @version 1.0
 ************************************************************************/
   class TLTimeFrequency : public TH2F {
   public:
      TLTimeFrequency (tftype_t tftype, 
                      const char* name, const char* title, 
                      Int_t nbinsx, Axis_t xlow, Axis_t xup, 
                      Int_t nbinsy, Axis_t ylow, Axis_t yup);
      /// Set plot type
      virtual void SetTfType (tftype_t tftype);
      /// Get plot type
      virtual tftype_t GetTfType() const {
         return fTfType; }
      /// Draws a time-frequency histogram (option is ignored)
      virtual void Draw (Option_t* option = 0);
      /// Paints a time-frequency histogram (used internally)
      virtual void Paint (Option_t* option);
      /// Add a new frequency series to the TF data object
      virtual void NewF (Float_t* y);
      /// Update (scroll) the time frequency histogram
      virtual void Update();
   
   protected:
      /// Time frequency plot type
      tftype_t		fTfType;
      /// Drawing options
      TString		fDrawOpt;
      /// Pointer to original canvas
      TVirtualPad*	fDrawCanv;
      /// Returns a valid canvas (or 0)
      virtual TVirtualPad* GetCanvas ();
   
      /// Returns a time-frequency painter
      static TVirtualHistPainter* GetTFPainter (TLTimeFrequency* tf);
   
    //ClassDef(TLTimeFrequency,0)  // Time frequency plot
   };


#endif //_LIGO_TLTFPainter
