// this class is for initializing and manipulating
// IIR filters implemented as second order sections
//
// Edward Daw, 20th May 2001
//   updated for greater efficiency and cds constructor 5th July 2002

#include "IIRSos.hh"
#include "Complex.hh"
#include <iostream>
#include <cmath>
#include "gds_fpclass.hh"
#include "constant.hh"

IIRSos::IIRSos(void) {
  _section_loaded = 0;
}
 
IIRSos::IIRSos( double gain, double c1, double c2, double c3, double c4) {
  init(gain, c1, c2, c3 ,c4);
}

IIRSos::IIRSos( double b0in, double b1in, double b2in,
		double a0in, double a1in, double a2in ) {
  init( b0in, b1in, b2in, a0in, a1in, a2in );
}

IIRSos::IIRSos( dComplex zeros, dComplex poles, double fs) {
  init(zeros, poles, fs);
}

IIRSos::IIRSos( double z0, double z1, dComplex poles, double fs) {
  init( z0, z1, poles, fs);
}

IIRSos::IIRSos( double z, dComplex poles, double fs) {
  init( z, poles, fs );
}

IIRSos::IIRSos( int nzeros, dComplex roots, double fs) {
  init( nzeros, roots, fs );
}

IIRSos::IIRSos( dComplex  zeros, double p0, double p1, double fs) {
  init( zeros, p0, p1, fs );
}

IIRSos::IIRSos( dComplex zeros, double p, double fs) {
  init( zeros, p, fs );
}

IIRSos::IIRSos( int nzeros, double z0, double z1, double p0, double p1, 
	       double fs) {
  init( nzeros, z0, z1, p0, p1, fs );
}

IIRSos::IIRSos( int nzeros, double r0, double r1, double r2, double fs) {
  init( nzeros, r0, r1, r2, fs );
}

IIRSos::IIRSos( int nzeros, double r0, double r1, double fs ) {
  init( nzeros, r0, r1, fs );
}

IIRSos::IIRSos( int nzeros, double r, double fs ) {
  init( nzeros, r, fs );
}

IIRSos::~IIRSos(void) {
}

int
IIRSos::init(double gain, double c1, double c2, double c3, double c4) {
  a1 = c1;
  a2 = c2;
  b0 = gain;
  b1 = c3 * gain;
  b2 = c4 * gain;
  _hzminusone = 0;
  _hzminustwo = 0;
  _section_loaded = 1;
  _period_known = false;
  return 0;
}

int
IIRSos::init(double b0in, double b1in, double b2in, 
	     double a0in, double a1in, double a2in  )
{
  b0 = b0in/a0in;
  b1 = b1in/a0in;
  b2 = b2in/a0in;
  a1 = a1in/a0in;
  a2 = a2in/a0in;
  _hzminusone = 0;
  _hzminustwo = 0;
  _section_loaded = 1;
  _period_known = false;
  return 0;
}

int
IIRSos::init(dComplex zero, dComplex pole, double fs) {
  double alpha(-Real(pole));
  double beta(Imag(pole));
  double gamma(-Real(zero));
  double delta(Imag(zero));
  _period = 1/fs;
  _period_known = true;
  // set filter coefficients
  double ssz(pow(gamma,2) + pow(delta,2));
  double ssp(pow(alpha,2) + pow(beta,2));
  double nm(4*fs*(fs+alpha)+ssp);
  a1=(2*(ssp-4*pow(fs,2)))/nm;
  a2=(4*fs*(fs-alpha)+ssp)/nm;
  b1=(2*(ssz-4*pow(fs,2)))/nm;
  b2=(4*fs*(fs-gamma)+ssz)/nm;
  b0=(4*fs*(fs+gamma)+ssz)/nm;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}

int
IIRSos::init(dComplex zeros, double p0, double p1, double fs) {
  double gamma(-Real(zeros));
  double delta(Imag(zeros));
  double ssz(pow(gamma,2)+pow(delta,2));
  double nm(0);
  _period=1/fs;  
  _period_known = true;
  _nzeros=2;
  double epsilon(-p0);
  double phi(-p1);
  _npoles=2;
  nm = ((2*fs+epsilon)*(2*fs+phi));
  a1 = (-2*(4*pow(fs,2)-epsilon*phi))/nm;
  a2 = ((2*fs-epsilon)*(2*fs-phi))/nm;
  b0 = (4*fs*(fs+gamma)+ssz)/nm;
  b1 = (2*(-4*pow(fs,2)+ssz))/nm;
  b2 = (4*fs*(fs-gamma)+ssz)/nm;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}

int
IIRSos::init(dComplex zeros, double p, double fs) {
  double gamma(-Real(zeros));
  double delta(Imag(zeros));
  double ssz(pow(gamma,2)+pow(delta,2));
  double nm(0);
  double epsilon(0);
  double phi(0);
  _period=1/fs;  
  _period_known = true;
  _nzeros=2;
  epsilon =(-p);
  std::cout << "warning: adding a pole at the Nyquist frequency to avoid"
	    << std::endl << "divergence of transfer function magnitude there."
	    << std::endl;
  phi = pi*fs;
  _npoles=2;
  nm = ((2*fs+epsilon)*(2*fs+phi));
  a1 = (-2*(4*pow(fs,2)-epsilon*phi))/nm;
  a2 = ((2*fs-epsilon)*(2*fs-phi))/nm;
  b0 = (phi/(2*pi))*(4*fs*(fs+gamma)+ssz)/nm;
  b1 = (phi/(2*pi))*(2*(-4*pow(fs,2)+ssz))/nm;
  b2 = (phi/(2*pi))*(4*fs*(fs-gamma)+ssz)/nm;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}
  
int
IIRSos::init(int nzeros, dComplex rootpair, double fs) {
  if( nzeros == 1 ) {
    dComplex zero(rootpair);
    double gamma(-Real(zero));
    double delta(Imag(zero));
    double ssz(pow(gamma,2)+pow(delta,2));
    double nm(0);
    double epsilon(0);
    double phi(0);
    _period=1/fs;  
    _period_known = true;
    _nzeros=2;
    std::cout << "warning: adding two poles at the Nyquist frequency to avoid"
	 << std::endl << "divergence of transfer function magnitude there."
	 << std::endl;
    epsilon = pi*fs;
    phi = pi*fs;
    _npoles=2;
    nm = ((2*fs+epsilon)*(2*fs+phi));
    a1 = (-2*(4*pow(fs,2)-epsilon*phi))/nm;
    a2 = ((2*fs-epsilon)*(2*fs-phi))/nm;
    b0 = epsilon*phi*(1/(4*pi*pi))*(4*fs*(fs+gamma)+ssz)/nm;
    b1 = epsilon*phi*(1/(4*pi*pi))*(2*(-4*pow(fs,2)+ssz))/nm;
    b2 = epsilon*phi*(1/(4*pi*pi))*(4*fs*(fs-gamma)+ssz)/nm;
    // set history buffers
    _hzminusone = 0;
    _hzminustwo = 0;
    // set filter loaded bit
    _section_loaded = 1;
  } else if ( nzeros == 0 ) {
    dComplex pole(rootpair);
    double alpha=(-Real(pole));
    double beta=(Imag(pole));
    double ssz(pow(alpha,2) + pow(beta,2));
    double nm(0);
    _period=1/fs;
    _period_known = true;
    _nzeros=2;
    nm = 4*fs*(fs+alpha)+ssz;
    _npoles=0;
    a1 = 2*(ssz-4*pow(fs,2))/nm;
    a2 = (4*fs*(fs-alpha)+ssz)/nm;
    b0 = 1/nm;
    b1 = 2/nm;
    b2 = 1/nm;
    // set history buffers
    _hzminusone = 0;
    _hzminustwo = 0;
    // set filter loaded bit
    _section_loaded = 1;
  } else {
    std::cerr << "ERROR: invalid initialization with 1 complex root pair"
	      << std::endl;
    return -1;
  }
  return 0;
}

int
IIRSos::init(double z, dComplex pole, double fs) {
  double alpha=(-Real(pole));
  double beta=(Imag(pole));
  double ssz(pow(alpha,2) + pow(beta,2));
  double nm(0);
  double mu(-z);
  _period=1/fs;
  _period_known = true;
  _nzeros=2;
  _npoles=1;
  nm = 4*fs*(fs+alpha)+ssz;
  a1 = 2*(ssz-4*pow(fs,2))/nm;
  a2 = (4*fs*(fs-alpha)+ssz)/nm;
  b0 = -(2*fs+mu)/nm;
  b1 = -2*mu/nm;
  b2 = (2*fs-mu)/nm;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}

int
IIRSos::init(double z0, double z1, dComplex pole, double fs) {
  double alpha=(-Real(pole));
  double beta=(Imag(pole));
  double ssz(pow(alpha,2) + pow(beta,2));
  double nm(0);
  double mu(0);
  double nu(0);
  _period=1/fs;
  _period_known = true;
  _nzeros=2;
  nm = 4*fs*(fs+alpha)+ssz;
  _npoles=2;
  mu = (-z0);
  nu = (-z1);
  a1 = 2*(ssz-4*pow(fs,2))/nm;
  a2 = (4*fs*(fs-alpha)+ssz)/nm;
  b0 = (4*pow(fs,2)+2*fs*(mu+nu)+mu*nu)/nm;
  b1 = 2*(nu*mu-4*pow(fs,2))/nm;
  b2 = (4*pow(fs,2)-2*fs*(mu+nu)+mu*nu)/nm;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}

int
IIRSos::init( int nzeros, double z0, double z1, double p0, double p1,
	      double fs ) {
  if( nzeros != 2 ) {
    std::cerr << "ERROR: should be two zeros and two poles with this "
	      << "init function." << std::endl;
    return -1;
  }
  double mu(0);
  double nu(0);
  double epsilon(0);
  double phi(0);
  double nm(0);
  _npoles = 2;
  _nzeros = 2;
  epsilon = (-p0);
  phi = (-p1);
  mu = (-z0);
  nu = (-z1);
  nm = 4*pow(fs,2)+2*fs*(epsilon+phi)+epsilon*phi;
  a1 = (2*(epsilon*phi-4*pow(fs,2)))/nm;
  a2 = (4*pow(fs,2)-2*fs*(epsilon+phi)+epsilon*phi)/nm;
  b0 = (4*pow(fs,2)+2*fs*(mu+nu)+mu*nu)/nm;
  b1 = (2*(mu*nu-4*pow(fs,2)))/nm;
  b2 = (4*pow(fs,2)-2*fs*(mu+nu)+mu*nu)/nm;
  // set period
  _period = 1/fs;
  _period_known = true;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
} 

int
IIRSos::init( int nzeros, double r0, double r1, double r2, double fs ) {
  double mu(0);
  double nu(0);
  double epsilon(0);
  double phi(0);
  double nm(0);
  if ( nzeros == 2 ) {
    double z0(r0);
    double z1(r1);
    double p(r2);
    _npoles = 2;
    _nzeros = 2;
    epsilon = (-p);
    std::cout << "WARNING: adding a real pole at the Nyquist frequency"
	      << std::endl;
    phi = pi*fs;
    mu = (-z0);
    nu = (-z1);
    nm = 4*pow(fs,2)+2*fs*(epsilon+phi)+epsilon*phi;
    a1 = (2*(epsilon*phi-4*pow(fs,2)))/nm;
    a2 = (4*pow(fs,2)-2*fs*(epsilon+phi)+epsilon*phi)/nm;
    b0 = (phi/(2*pi))*(4*pow(fs,2)+2*fs*(mu+nu)+mu*nu)/nm;
    b1 = (phi/(2*pi))*(2*(mu*nu-4*pow(fs,2)))/nm;
    b2 = (phi/(2*pi))*(4*pow(fs,2)-2*fs*(mu+nu)+mu*nu)/nm;
  } else if ( nzeros == 1 ) {
    double z(r0);
    double p0(r1);
    double p1(r2);
    _npoles = 2;
    _nzeros = 1;
    mu = (-z);
    epsilon = (-p0);
    phi = (-p1);
    nm = 4*pow(fs,2)+2*fs*(epsilon+phi)+epsilon*phi;
    a1 = (2*(epsilon*phi-4*pow(fs,2)))/nm;
    a2 = (4*pow(fs,2)-2*fs*(epsilon+phi)+epsilon*phi)/nm;
    b0 = -(2*fs+mu)/nm;
    b1 = -2*mu/nm;
    b2 = (2*fs-mu)/nm;
  } else {
    std::cerr << "ERROR: three real root init function inappropriate."
	      << std::endl;
    return -1;
  }
  // set period
  _period = 1/fs;
  _period_known = true;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}

int 
IIRSos::init( int nzeros, double r0, double r1, double fs ) {
  double mu(0);
  double nu(0);
  double epsilon(0);
  double phi(0);
  double nm(0);
  if ( nzeros == 2) {
    double z0(r0);
    double z1(r1);
    _npoles = 2;
    _nzeros = 2;
    std::cout << "warning: adding two real poles at the Nyquist frequency"
	      << std::endl;
    epsilon = pi*fs;
    phi = pi*fs;
    mu = (-z0);
    nu = (-z1);
    nm = 4*pow(fs,2)+2*fs*(epsilon+phi)+epsilon*phi;
    a1 = (2*(epsilon*phi-4*pow(fs,2)))/nm;
    a2 = (4*pow(fs,2)-2*fs*(epsilon+phi)+epsilon*phi)/nm;
    b0 = epsilon*phi*(1/(4*pi*pi))*(4*pow(fs,2)+2*fs*(mu+nu)+mu*nu)/nm;
    b1 = epsilon*phi*(1/(4*pi*pi))*(2*(mu*nu-4*pow(fs,2)))/nm;
    b2 = epsilon*phi*(1/(4*pi*pi))*(4*pow(fs,2)-2*fs*(mu+nu)+mu*nu)/nm;
  } else if ( nzeros == 1 ) {
    _npoles = 1;
    _nzeros = 1;    
    double z(r0);
    double p(r1);
    epsilon = (-p);
    mu = (-z);
    nm = -(2*fs+epsilon);
    a1 = (2*fs-epsilon)/nm;
    a2 = 0;
    b0 = -(2*fs+mu)/nm;
    b1 = (2*fs-mu)/nm;
    b2 = 0;
  } else if ( nzeros == 0 ) {
    double p0(r0);
    double p1(r1);
    _npoles = 2;
    _nzeros = 0;    
    epsilon = (-p0);
    phi = (-p1);
    nm = 4*pow(fs,2)+2*fs*(epsilon+phi)+epsilon*phi;
    a1 = (2*(epsilon*phi-4*pow(fs,2)))/nm;
    a2 = (4*pow(fs,2)-2*fs*(epsilon+phi)+epsilon*phi)/nm;
    b0 = 1/nm;
    b1 = 2/nm;
    b2 = 1/nm;
  } else {
    std::cerr << "ERROR: two real root init fuction inappropriate"
	      << std::endl;
    return -1;
  }
  // set period
  _period = 1/fs;
  _period_known = true;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}

int
IIRSos::init( int nzeros, double r, double fs ) {
  if ( nzeros == 1 ) {
    double z(r);
    _npoles = 1;
    _nzeros = 1;    
    std::cout << "warning: inserting extra pole at Nyquist frequency"
	      << std::endl;
    double epsilon(pi*fs);
    double mu(-z);
    double nm(-(2*fs+epsilon));
    a1 = (2*fs-epsilon)/nm;
    a2 = 0;
    b0 = (epsilon/(2*pi))*(-(2*fs+mu)/nm);
    b1 = (epsilon/(2*pi))*(2*fs-mu)/nm;
    b2 = 0;
  } else if ( nzeros == 0 ) {
    double p(r);
    _npoles = 1;
    _nzeros = 0;
    double epsilon(-p);
    double nm(-(2*fs+epsilon));
    a1 = (2*fs-epsilon)/nm;
    a2 = 0;
    b0 = 1/nm;
    b1 = 1/nm;
    b2 = 0;
  } else {
    std::cerr << "ERROR: one real root init fuction inappropriate"
	      << std::endl;
    return -1;
  }
  // set period
  _period = 1/fs;
  _period_known = true;
  // set history buffers
  _hzminusone = 0;
  _hzminustwo = 0;
  // set filter loaded bit
  _section_loaded = 1;
  return 0;
}

int
IIRSos::dump(std::ostream& output) const {
  output << "b0 = " << b0 << std::endl
	 << "b1 = " << b1 << std::endl
	 << "b2 = " << b2 << std::endl
	 << "al = " << a1 << std::endl
	 << "a2 = " << a2 << std::endl;
  return 0;
}

int
IIRSos::apply(double data_in, double* data_out) {

  // apply single second order section
  double hzzero = data_in - _hzminusone * a1 - _hzminustwo * a2;
  *data_out = hzzero * b0 + _hzminusone * b1 + _hzminustwo * b2;

  // update history buffers
  _hzminustwo = _hzminusone;
  _hzminusone = hzzero;

  // return
  return 0;
}

int 
IIRSos::apply(double* pdata, int ndata) {

  register double hzzero, hzm1(_hzminusone), hzm2(_hzminustwo);
  register double aa1(a1), aa2(a2);
  register double bb0(b0), bb1(b1), bb2(b2);
  double* p(pdata);
  for( int i=0; i<ndata; ++i ) {

  // apply single second order section

    hzzero = *p - hzm1 * aa1 - hzm2 * aa2;
    *p++ = hzzero * bb0 + hzm1 * bb1 + hzm2 * bb2;
    
    // update history buffers
    hzm2 = hzm1;
    hzm1 = hzzero;

  }
  _hzminusone = hzm1;
  _hzminustwo = hzm2;
  return 0;
}

int 
IIRSos::apply(float* pdata, int ndata) {

  register double hzzero, hzm1(_hzminusone), hzm2(_hzminustwo);
  register double aa1(a1), aa2(a2);
  register double bb0(b0), bb1(b1), bb2(b2);
  float* p(pdata);
  for( int i=0; i<ndata; ++i ) {

  // apply single second order section

    hzzero = double(*p) - hzm1 * aa1 - hzm2 * aa2;
    *p++ = float(hzzero * bb0 + hzm1 * bb1 + hzm2 * bb2);
    
    // update history buffers
    hzm2 = hzm1;
    hzm1 = hzzero;

  }
  _hzminusone = hzm1;
  _hzminustwo = hzm2;
  return 0;
}

void
IIRSos::reset(void) {
  _hzminusone = 0;
  _hzminustwo = 0;
}

bool
IIRSos::test(void) const {
    return !isfinite(_hzminusone) || !isfinite(_hzminustwo);
}

dComplex
IIRSos::H(const dComplex& iz) const {
  dComplex iz2 = iz*iz; 
  return (dComplex(b0,0)+b1*iz+b2*iz2) / (dComplex(1,0)+a1*iz+a2*iz2);
  //return ((iz*b2 + dComplex(b1))*iz + dComplex(b0)) / 
  //       ((iz*a2 + dComplex(a1))*iz + dComplex(1.0));
}
